﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Runtime.Loader;
using System.Runtime.Versioning;
using System.Threading.Tasks;

namespace HeuristicLab.PluginInfrastructure.Runner {

  public class AssemblyLoader : IAssemblyLoader {
    private class AssemblyNameComparer : IEqualityComparer<AssemblyName> {
      private object concatProps(AssemblyName name) => name.FullName + name.Version.ToString();
      public bool Equals(AssemblyName lhs, AssemblyName rhs) => concatProps(lhs).Equals(concatProps(rhs));
      public int GetHashCode(AssemblyName obj) => concatProps(obj).GetHashCode();
    }

    //https://github.com/dotnet/coreclr/blob/master/Documentation/design-docs/assemblyloadcontext.md
    private class ALC : AssemblyLoadContext {
      private Dictionary<AssemblyName, Assembly> loadedAssemblies = new Dictionary<AssemblyName, Assembly>(new AssemblyNameComparer());
      private string BasePath { get; set; }
      public ALC(string basePath) { BasePath = basePath; }
      protected override Assembly Load(AssemblyName assemblyName) {
        if (loadedAssemblies.ContainsKey(assemblyName)) return loadedAssemblies[assemblyName];

        string path = $"{BasePath}{Path.DirectorySeparatorChar}{assemblyName.Name}.dll";
        Assembly asm = null;
        if (File.Exists(path))
          asm = Assembly.LoadFrom(path);
        else 
          asm = Assembly.Load(assemblyName);
        loadedAssemblies.Add(assemblyName, asm);
        foreach (var a in asm.GetReferencedAssemblies())
          Load(a);
        return asm;
      }
    }

    private ALC alc;
    public string BasePath { get; set; }
    public string AssemblyPrefix { get; set; }

    //public IList<Assembly> Assemblies { private set; get; } = new List<Assembly>();
    //public IList<Type> Types { private set; get; } = new List<Type>();

    private IList<Assembly> assemblies = new List<Assembly>();
    public IEnumerable<Assembly> Assemblies => assemblies;

    private IList<Type> types = new List<Type>();
    public IEnumerable<Type> Types => types;

    public AssemblyLoader() : this(Directory.GetCurrentDirectory()) { }
    public AssemblyLoader(string basePath) {
      this.BasePath = basePath;
      alc = new ALC(BasePath);
      /*
      ScanTypesInBasePath();
      Types = GetTypes();
      */
      /*
      IEnumerable<Assembly> asms = CheckDependencies(GetLoadableAssemblies(GetAssembliesFromBasePath()));
      LoadAssemblies(asms);
      LoadTypes(asms);
      */
      //IEnumerable<Assembly> asms = GetLoadableAssemblies(GetAssembliesFromBasePath());
      //LoadAssemblies(asms);
      //LoadTypes(asms);
    }

    // 1.
    private IEnumerable<string> GetAssembliesFromBasePath() {
      return Directory.GetFiles(BasePath, $"*.dll", SearchOption.AllDirectories);
    }
    /*
    // 2.
    private IEnumerable<Assembly> GetLoadableAssemblies(IEnumerable<string> paths) {
      IList<Assembly> assemblies = new List<Assembly>();      
      foreach (string path in paths) {
        try {
          var asm = Assembly.LoadFile(path);
          if (asm != null)
            assemblies.Add(asm);
          else
            Console.WriteLine($"cannot load assembly with path={path}");
        } catch (Exception e) {
          Console.WriteLine($"exception occured for assembly with path={path}, type of exception={e.GetType()}");
        }
      }
      
      return assemblies;
    }
    
    // 3.
    private IEnumerable<Assembly> CheckDependencies(IEnumerable<Assembly> assemblies) {
      IEqualityComparer<AssemblyName> comp = new AssemblyNameComparer();
      IDictionary<AssemblyName, Assembly> dict = new Dictionary<AssemblyName, Assembly>(comp);

      // fill dict
      foreach (Assembly a in assemblies)
        dict.Add(a.GetName(), a);

      Parallel.ForEach(new List<Assembly>(dict.Values), asm => {
        if (CheckDependencyRecursiveHelper(asm.GetName(), asm.GetReferencedAssemblies(), comp, dict))
          lock(dict)
            dict.Remove(asm.GetName());
      });
      return dict.Values;
    }

    private bool CheckDependencyRecursiveHelper(AssemblyName toCheck, IEnumerable<AssemblyName> refs, IEqualityComparer<AssemblyName> comp, IDictionary<AssemblyName, Assembly> dict) {
      foreach(AssemblyName name in refs) {
        if (comp.Equals(toCheck, name) || !dict.ContainsKey(name) || CheckDependencyRecursiveHelper(toCheck, dict[name].GetReferencedAssemblies(), comp, dict)) return true;
      }
      return false;
    }
    
    // 4.
    private void LoadAssemblies(IEnumerable<Assembly> assemblies) {

      foreach (Assembly asm in assemblies) {
        try {
          Assemblies.Add(alc.LoadFromAssemblyName(asm.GetName()));
        } catch(Exception e) { }
      }
    }
    */
    private void LoadTypes(IEnumerable<Assembly> assemblies) {
      foreach (Assembly asm in assemblies) {
        try {
          foreach (Type t in asm.GetExportedTypes()) {
            types.Add(t);
          }
        } catch (ReflectionTypeLoadException e) {
          try {
            foreach (Type t in e.Types) {
              if (t != null) {
                types.Add(t);
              }
            }
          } catch (BadImageFormatException) { }

        } catch (Exception e) {
          Tracing.Logger.Error(
            $"Exception occured while loading types of assembly {asm.FullName}! \n " +
            $"---- Stacktrace ---- \n {e}");
        }
      }
    }
    /*
    public Type GetType(string type) {
      foreach (Assembly asm in Assemblies) {
        Type t = asm.GetType(type);
        if (t != null) return t;
      }
      return null;
    }
    */

    public IEnumerable<Assembly> GetReflectionOnlyAssemblies() {
      IList<Assembly> list = new List<Assembly>();
      foreach (string path in GetAssembliesFromBasePath()) {
        try {
          var asm = Assembly.LoadFrom(path);
          if (asm != null)
            list.Add(asm);
          else
            Tracing.Logger.Error($"Unnable to load assembly with path {path}!");
        } catch (Exception e) {
          Tracing.Logger.Error(
            $"Exception occured while loading assembly from path {path}! \n " +
            $"---- Stacktrace ---- \n {e}");
        }
      }
      return list;
    }

    public IEnumerable<Assembly> LoadAssemblies(IEnumerable<AssemblyName> assemblyNames) {
      foreach (AssemblyName asm in assemblyNames) {
        try {
          this.assemblies.Add(alc.LoadFromAssemblyName(asm));
        } catch (Exception e) {
          Tracing.Logger.Error(
            $"Exception occured while loading assembly: {asm.FullName}! \n " +
            $"---- Stacktrace ---- \n {e}");
        }
      }
      LoadTypes(this.Assemblies);
      return this.Assemblies;
    }
  }
}